"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.ind_bollinger_bands = exports.ind_alligator = exports.indi_rsi = exports.indi_obv = exports.indi_macd = exports.indi_adx = void 0;
const feature_statistics_1 = require("./feature_statistics");
const feature_ohlc_1 = require("./feature_ohlc");
const lodash_round_1 = __importDefault(require("lodash.round"));
const indi_adx = (ohlc_data, period) => {
    const trueRanges = [];
    const positiveDMs = [];
    const negativeDMs = [];
    const dxValues = [];
    const adxValues = [];
    const sanitized_ohlc = (0, feature_ohlc_1.sanitize_ohlc)(ohlc_data);
    // Calculate True Range (TR), Positive Directional Movement (+DM), Negative Directional Movement (-DM)
    for (let i = 1; i < sanitized_ohlc.length; i++) {
        const currentHigh = sanitized_ohlc[i].high;
        const currentLow = sanitized_ohlc[i].low;
        const previousClose = sanitized_ohlc[i - 1].close;
        const trueRange = Math.max(currentHigh - currentLow, Math.abs(currentHigh - previousClose), Math.abs(currentLow - previousClose));
        trueRanges.push(trueRange);
        const positiveDM = currentHigh - sanitized_ohlc[i - 1].high;
        const negativeDM = sanitized_ohlc[i - 1].low - currentLow;
        positiveDMs.push(positiveDM > 0 ? positiveDM : 0);
        negativeDMs.push(negativeDM > 0 ? negativeDM : 0);
    }
    // Calculate the smoothed average of True Range (ATR)
    const atrValues = calculateSmoothedAverage(trueRanges, period);
    // Calculate the smoothed averages of Positive Directional Movement (+DM) and Negative Directional Movement (-DM)
    const positiveDMMAs = calculateSmoothedAverage(positiveDMs, period);
    const negativeDMMAs = calculateSmoothedAverage(negativeDMs, period);
    // Calculate the Directional Movement Index (DX) and ADX
    for (let i = period; i < sanitized_ohlc.length - 1; i++) {
        const diPlus = (positiveDMMAs[i - period] / atrValues[i - period]) * 100;
        const diMinus = (negativeDMMAs[i - period] / atrValues[i - period]) * 100;
        const dx = (Math.abs(diPlus - diMinus) / (diPlus + diMinus)) * 100;
        dxValues.push(dx);
        if (dxValues.length >= period) {
            const adx = calculateSmoothedAverage(dxValues, period)[dxValues.length - period];
            adxValues.push(adx);
        }
    }
    return adxValues;
};
exports.indi_adx = indi_adx;
const indi_macd = (data, shortPeriod = 12, longPeriod = 26, signalPeriod = 9) => {
    // Calculate the short EMA
    const shortEMA = (0, feature_statistics_1.stats_ema)(data, shortPeriod);
    // Calculate the long EMA
    const longEMA = (0, feature_statistics_1.stats_ema)(data, longPeriod);
    // Calculate the MACD line
    const macdLine = [];
    for (let i = 0; i < data.length; i++) {
        macdLine.push(shortEMA[i] - longEMA[i]);
    }
    // Calculate the signal line using SMA of the MACD line
    const signalLine = (0, feature_statistics_1.stats_ema)(macdLine, signalPeriod);
    // Return the MACD line and signal line
    return {
        macd: macdLine,
        signal_line: signalLine,
    };
};
exports.indi_macd = indi_macd;
const indi_obv = (data) => {
    const obvValues = [];
    obvValues.push(0); // Initialize OBV with zero as the starting value
    for (let i = 1; i < data.length; i++) {
        const currentClose = data[i].close;
        const previousClose = data[i - 1].close;
        const volume = parseInt(data[i].volume.toString());
        if (currentClose > previousClose) {
            // Add the volume to OBV if the current close price is higher than the previous close price
            obvValues.push(obvValues[i - 1] + volume);
        }
        else if (currentClose < previousClose) {
            // Subtract the volume from OBV if the current close price is lower than the previous close price
            obvValues.push(obvValues[i - 1] - volume);
        }
        else {
            // If the current close price is equal to the previous close price, keep the OBV value unchanged
            obvValues.push(obvValues[i - 1]);
        }
    }
    return obvValues;
};
exports.indi_obv = indi_obv;
const indi_rsi = (data, period = 7) => {
    const gains = [];
    const losses = [];
    const rsi = [];
    // Calculate the price changes and separate gains and losses
    for (let i = 1; i < data.length; i++) {
        const change = data[i] - data[i - 1];
        if (change >= 0) {
            gains.push(change);
            losses.push(0);
        }
        else {
            gains.push(0);
            losses.push(-change);
        }
    }
    // Calculate the average gains and losses for the first period
    let avgGain = (0, feature_statistics_1.stats_mean)(gains.slice(0, period));
    let avgLoss = (0, feature_statistics_1.stats_mean)(losses.slice(0, period));
    // Calculate the initial RSI value
    let rs = avgGain / avgLoss;
    const initialRSI = 100 - 100 / (1 + rs);
    rsi.push(initialRSI);
    // Calculate RSI for the remaining data
    for (let i = period; i < data.length - 1; i++) {
        const currentGain = gains[i];
        const currentLoss = losses[i];
        // Smooth the average gains and losses using the previous averages
        avgGain = (avgGain * (period - 1) + currentGain) / period;
        avgLoss = (avgLoss * (period - 1) + currentLoss) / period;
        rs = avgGain / avgLoss;
        const currentRSI = 100 - 100 / (1 + rs);
        rsi.push(currentRSI);
    }
    return rsi;
};
exports.indi_rsi = indi_rsi;
const ind_alligator = (ohlc_data, jawPeriod, teethPeriod, lipsPeriod) => {
    const jawValues = [];
    const teethValues = [];
    const lipsValues = [];
    const sanitized_ohlc = (0, feature_ohlc_1.sanitize_ohlc)(ohlc_data);
    for (let i = jawPeriod - 1; i < sanitized_ohlc.length; i++) {
        const jawStartIndex = i - jawPeriod + 1;
        const teethStartIndex = i - teethPeriod + 1;
        const lipsStartIndex = i - lipsPeriod + 1;
        const jawHighs = sanitized_ohlc.slice(jawStartIndex, i + 1).map((d) => d.high);
        const jawLowestHigh = Math.min(...jawHighs);
        jawValues.push(jawLowestHigh);
        const teethHighs = sanitized_ohlc.slice(teethStartIndex, i + 1).map((d) => d.high);
        const teethLowestHigh = Math.min(...teethHighs);
        teethValues.push(teethLowestHigh);
        const lipsHighs = sanitized_ohlc.slice(lipsStartIndex, i + 1).map((d) => d.high);
        const lipsLowestHigh = Math.min(...lipsHighs);
        lipsValues.push(lipsLowestHigh);
    }
    return {
        jaw: jawValues,
        teeth: teethValues,
        lips: lipsValues,
    };
};
exports.ind_alligator = ind_alligator;
const ind_bollinger_bands = (data, period = 20, multiplier = 2) => {
    const closePrices = data.map(d => (0, lodash_round_1.default)(Number(d.close), 2));
    const middle = [];
    const upper = [];
    const lower = [];
    function simpleMovingAverage(prices, period, index) {
        const slice = prices.slice(index - period, index);
        const sum = slice.reduce((acc, price) => acc + price, 0);
        return sum / period;
    }
    function standardDeviation(prices, period, index, sma) {
        const slice = prices.slice(index - period, index);
        const variance = slice.reduce((acc, price) => acc + Math.pow(price - sma, 2), 0) / period;
        return Math.sqrt(variance);
    }
    for (let i = period; i <= closePrices.length; i++) {
        const sma = simpleMovingAverage(closePrices, period, i);
        const stdDev = standardDeviation(closePrices, period, i, sma);
        middle.push(sma);
        upper.push(sma + multiplier * stdDev);
        lower.push(sma - multiplier * stdDev);
    }
    return { middle, upper, lower };
};
exports.ind_bollinger_bands = ind_bollinger_bands;
const calculateSmoothedAverage = (data, period) => {
    const smoothedData = [];
    const sum = data.slice(0, period).reduce((acc, val) => acc + val, 0);
    const average = sum / period;
    smoothedData.push(average);
    for (let i = period; i < data.length; i++) {
        const smoothedValue = (smoothedData[i - period] * (period - 1) + data[i]) / period;
        smoothedData.push(smoothedValue);
    }
    return smoothedData;
};
//# sourceMappingURL=feature_indicators.js.map